/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * httpd.c: simple http daemon for answering WWW file requests
 *
 * 
 * 03-21-93  Rob McCool wrote original code (up to NCSA HTTPd 1.3)
 * 
 * 03-06-95  blong
 *  changed server number for child-alone processes to 0 and changed name
 *   of processes
 *
 * 03-10-95  blong
 *      Added numerous speed hacks proposed by Robert S. Thau (rst@ai.mit.edu) 
 *      including set group before fork, and call gettime before to fork
 *      to set up libraries.
 *
 * 04-14-95  rst / rh
 *      Brandon's code snarfed from NCSA 1.4, but tinkered to work with the
 *      Apache server, and also to have child processes do accept() directly.
 *
 * April-July '95 rst
 *      Extensive rework for Apache.
 */

#define CORE_PRIVATE

#include "httpd.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"	/* for read_config */
#include "http_protocol.h"	/* for read_request */
#include "http_request.h"	/* for process_request */
#include "http_conf_globals.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_vhost.h"
#include "util_script.h"	/* to force util_script.c linking */
#include "util_uri.h"
#include "scoreboard.h"

#include <prinit.h>
#include <prthread.h>
#include <prnetdb.h>
#include <prlock.h>

#include <setjmp.h>

/* This next function is never used. It is here to ensure that if we
 * make all the modules into shared libraries that core httpd still
 * includes the full Apache API. Without this function the objects in
 * main/util_script.c would not be linked into a minimal httpd.
 * And the extra prototype is to make gcc -Wmissing-prototypes quiet.
 */
extern void ap_force_library_loading(void);
void ap_force_library_loading(void) {
    ap_add_cgi_vars(NULL);
}

#include "explain.h"

#if !defined(max)
#define max(a,b)        (a > b ? a : b)
#endif

DEF_Explain

/* Defining GPROF when compiling uses the moncontrol() function to
 * disable gprof profiling in the parent, and enable it only for
 * request processing in children (or in one_process mode).  It's
 * absolutely required to get useful gprof results under linux
 * because the profile itimers and such are disabled across a
 * fork().  It's probably useful elsewhere as well.
 */
#ifdef GPROF
extern void moncontrol(int);
#define MONCONTROL(x) moncontrol(x)
#else
#define MONCONTROL(x)
#endif

/*
 * Actual definitions of config globals... here because this is
 * for the most part the only code that acts on 'em.  (Hmmm... mod_main.c?)
 */

int ap_standalone;
uid_t ap_user_id;
char *ap_user_name;
gid_t ap_group_id;
#ifdef MULTIPLE_GROUPS
gid_t group_id_list[NGROUPS_MAX];
#endif
int ap_max_requests_per_child;
int ap_threads_per_child;
int ap_excess_requests_per_child;
char *ap_pid_fname;
char *ap_scoreboard_fname;
char *ap_lock_fname;
char *ap_server_argv0;
PRNetAddr ap_bind_address;
int ap_daemons_to_start;
int ap_daemons_min_free;
int ap_daemons_max_free;
int ap_daemons_limit;
PRTime ap_restart_time;
int ap_suexec_enabled = 0;
int ap_listenbacklog;
int ap_dump_settings;

/*
 * The max child slot ever assigned, preserved across restarts.  Necessary
 * to deal with MaxClients changes across SIGUSR1 restarts.  We use this
 * value to optimize routines that have to scan the entire scoreboard.
 */
static int max_daemons_limit = -1;

/*
 * During config time, listeners is treated as a NULL-terminated list.
 * child_main previously would start at the beginning of the list each time
 * through the loop, so a socket early on in the list could easily starve out
 * sockets later on in the list.  The solution is to start at the listener
 * after the last one processed.  But to do that fast/easily in child_main it's
 * way more convenient for listeners to be a ring that loops back on itself.
 * The routine setup_listeners() is called after config time to both open up
 * the sockets and to turn the NULL-terminated list into a ring that loops back
 * on itself.
 *
 * head_listener (in child_main) is used by each child to keep track of what
 * they consider to be the "start" of the ring.  It is also set by
 * make_child to ensure that new children also don't starve any sockets.
 *
 * Note that listeners != NULL is ensured by read_config().
 */
listen_rec *ap_listeners;

API_VAR_EXPORT char ap_server_root[MAX_STRING_LEN];
char ap_server_confname[MAX_STRING_LEN];
char ap_coredump_dir[MAX_STRING_LEN];

array_header *ap_server_pre_read_config;
array_header *ap_server_post_read_config;

/* stuff that needs thread local store in main */
typedef struct {
    /* NSPR doesn't have any method for a thread to exit, except for it to
     * return from its starter function... this jmp_buf is our way out.
     */
    jmp_buf thread_exit;
    jmp_buf jmpbuffer;
    int volatile exit_after_unblock;
    int volatile alarms_blocked;
    int volatile alarm_pending;
    void (*volatile alarm_fn) (int);
    pool *pchild;		/* Pool for httpd child stuff */
    int my_child_num;
    int volatile usr1_just_die;
    int volatile deferred_die;
    int requests_this_child;
} tls_main_t;

static PRUintn tls_main_key;
#define tls()	((tls_main_t *)PR_GetThreadPrivate(tls_main_key))

/* used by accept_loop(), which is serialized */
static PRLock *accept_loop_lock;
static listen_rec *head_listener;

/* *Non*-shared http_main globals... */

static server_rec *server_conf;
static PRPollDesc *listenfds;
static PRIntn num_listenfds;
static pid_t pgrp;

/* one_process --- debugging mode variable; can be set from the command line
 * with the -X flag.  If set, this gets you the child_main loop running
 * in the process which originally started up (no detach, no make_child),
 * which is a pretty nice debugging environment.  (You'll get a SIGHUP
 * early in standalone_main; just continue through.  This is the server
 * trying to kill off any child processes which it might have lying
 * around --- Apache doesn't keep track of their pids, it just sends
 * SIGHUP to the process group, ignoring it in the root process.
 * Continue through and you'll be fine.).
 */

static int one_process = 0;

/* set if timeouts are to be handled by the children and not by the parent.
 * i.e. child_timeouts = !standalone || one_process.
 */
static int child_timeouts;

#ifdef DEBUG_SIGSTOP
int raise_sigstop_flags;
#endif

#ifndef NO_OTHER_CHILD
/* used to maintain list of children which aren't part of the scoreboard */
typedef struct other_child_rec other_child_rec;
struct other_child_rec {
    other_child_rec *next;
    int pid;
    void (*maintenance) (int, void *, int);
    void *data;
    int write_fd;
};
static other_child_rec *other_children;
#endif

static pool *pperm;		/* Pool for permanent stuff */
static pool *pconf;		/* Pool for config stuff */
static pool *ptemp;		/* Pool for temporary config stuff */
static pool *pcommands;	/* Pool for -C and -c switches */

scoreboard *ap_scoreboard_image = NULL;

/* a clean exit from a child with proper cleanup */
static void __attribute__((noreturn)) clean_child_exit(int code)
{
    if (tls()->pchild) {
	ap_child_exit_modules(tls()->pchild, server_conf);
	ap_destroy_pool(tls()->pchild);
    }
    longjmp(tls()->thread_exit, 1);
}

/* Default --- no serialization.  Other methods *could* go here,
 * as #elifs...
 */
/* Multithreaded systems don't complete between processes for
 * the sockets. */
#define NO_SERIALIZED_ACCEPT
#define accept_mutex_child_init(x)
#define accept_mutex_init(x)
#define accept_mutex_on()
#define accept_mutex_off()

/* On some architectures it's safe to do unserialized accept()s in the single
 * Listen case.  But it's never safe to do it in the case where there's
 * multiple Listen statements.  Define SINGLE_LISTEN_UNSERIALIZED_ACCEPT
 * when it's safe in the single Listen case.
 */
#ifdef SINGLE_LISTEN_UNSERIALIZED_ACCEPT
#define SAFE_ACCEPT(stmt) do {if(ap_listeners->next != ap_listeners) {stmt;}} while(0)
#else
#define SAFE_ACCEPT(stmt) do {stmt;} while(0)
#endif

static void usage(char *bin)
{
    char pad[MAX_STRING_LEN];
    int i;

    for (i = 0; i < strlen(bin); i++)
	pad[i] = ' ';
    pad[i] = '\0';
#ifdef SHARED_CORE
    PR_fprintf(PR_STDERR, "Usage: %s [-L directory] [-d directory] [-f file]\n", bin);
#else
    PR_fprintf(PR_STDERR, "Usage: %s [-d directory] [-f file]\n", bin);
#endif
    PR_fprintf(PR_STDERR, "       %s [-C \"directive\"] [-c \"directive\"]\n", pad);
    PR_fprintf(PR_STDERR, "       %s [-v] [-V] [-h] [-l] [-S]\n", pad);
    PR_fprintf(PR_STDERR, "Options:\n");
#ifdef SHARED_CORE
    PR_fprintf(PR_STDERR, "  -L directory     : specify an alternate location for shared object files\n");
#endif
    PR_fprintf(PR_STDERR, "  -d directory     : specify an alternate initial ServerRoot\n");
    PR_fprintf(PR_STDERR, "  -f file          : specify an alternate ServerConfigFile\n");
    PR_fprintf(PR_STDERR, "  -C \"directive\"   : process directive before reading config files\n");
    PR_fprintf(PR_STDERR, "  -c \"directive\"   : process directive after  reading config files\n");
    PR_fprintf(PR_STDERR, "  -v               : show version number\n");
    PR_fprintf(PR_STDERR, "  -V               : show compile settings\n");
    PR_fprintf(PR_STDERR, "  -h               : list available configuration directives\n");
    PR_fprintf(PR_STDERR, "  -l               : list compiled-in modules\n");
    PR_fprintf(PR_STDERR, "  -S               : show parsed settings (currently only vhost settings)\n");
    exit(1);
}


/*
 * More machine-dependent networking gooo... on some systems,
 * you've got to be *really* sure that all the packets are acknowledged
 * before closing the connection, since the client will not be able
 * to see the last response if their TCP buffer is flushed by a RST
 * packet from us, which is what the server's TCP stack will send
 * if it receives any request data after closing the connection.
 *
 * In an ideal world, this function would be accomplished by simply
 * setting the socket option SO_LINGER and handling it within the
 * server's TCP stack while the process continues on to the next request.
 * Unfortunately, it seems that most (if not all) operating systems
 * block the server process on close() when SO_LINGER is used.
 * For those that don't, see USE_SO_LINGER below.  For the rest,
 * we have created a home-brew lingering_close.
 *
 * Many operating systems tend to block, puke, or otherwise mishandle
 * calls to shutdown only half of the connection.  You should define
 * NO_LINGCLOSE in conf.h if such is the case for your system.
 */
#ifndef MAX_SECS_TO_LINGER
#define MAX_SECS_TO_LINGER 30
#endif

#ifdef USE_SO_LINGER
#define NO_LINGCLOSE		/* The two lingering options are exclusive */

static void sock_enable_linger(PRFileDesc *s)
{
    PRSocketOptionData sod;
    
    sod.option = PR_SockOpt_Linger;
    sod.value.linger.polarity = 1;
    sod.value.linger.linger = PRSecondsToInterval(MAX_SECS_TO_LINGER);

    if (PRSetSockOpt(s, &sod) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf,
	            "setsockopt: (SO_LINGER)");
	/* not a fatal error */
    }
}

#else
#define sock_enable_linger(s)	/* NOOP */
#endif /* USE_SO_LINGER */

#ifndef NO_LINGCLOSE

/* Since many clients will abort a connection instead of closing it,
 * attempting to log an error message from this routine will only
 * confuse the webmaster.  There doesn't seem to be any portable way to
 * distinguish between a dropped connection and something that might be
 * worth logging.
 */
static void lingering_close(request_rec *r)
{
    char dummybuf[IOBUFSIZE];
    PRPollDesc pd;
    PRFileDesc *lsd;
    PRIntervalTime poll_timer;
    PRIntervalTime epoch;
    PRIntervalTime max_wait;

    /* Prevent a slow-drip client from holding us here indefinitely */

    epoch = PR_IntervalNow();
    max_wait = PR_SecondsToInterval(MAX_SECS_TO_LINGER);
    ap_bsetopt(r->connection->client, BO_TIMEOUT, &max_wait);

    /* Send any leftover data to the client, but never try to again */

    if (ap_bflush(r->connection->client) == -1) {
	ap_bclose(r->connection->client);
	return;
    }
    ap_bsetflag(r->connection->client, B_EOUT, 1);

    /* Close our half of the connection --- send the client a FIN */

    lsd = r->connection->client->fd;

    if (PR_Shutdown(lsd, PR_SHUTDOWN_SEND) != PR_SUCCESS 
	|| ap_is_aborted(r->connection)) {
	ap_bclose(r->connection->client);
	return;
    }

    /* Set up to wait for readable data on socket... */
    pd.fd = lsd;
    pd.in_flags = PR_POLL_READ;

    /* Wait for readable data or error condition on socket;
     * slurp up any data that arrives...  We exit when we go for 
     * an interval of tv length without getting any more data, get an
     * error from PR_Poll(), get an exception on lsd, get an error or EOF
     * on a read, or the timer expires.
     */
    /* We use a 2 second timeout because current (Feb 97) browsers
     * fail to close a connection after the server closes it.  Thus,
     * to avoid keeping the child busy, we are only lingering long enough
     * for a client that is actively sending data on a connection.
     * This should be sufficient unless the connection is massively
     * losing packets, in which case we might have missed the RST anyway.
     */
    poll_timer = PR_SecondsToInterval(2);
    do {
	pd.out_flags = 0;
    } while (PR_Poll(&pd, 1, poll_timer) == 1
	    && PR_Read(lsd, dummybuf, sizeof(dummybuf)) > 0
	    && (PRIntervalTime)(PR_IntervalNow() - epoch) < max_wait);

    /* Should now have seen final ack.  Safe to finally kill socket */

    ap_bclose(r->connection->client);
}
#endif /* ndef NO_LINGCLOSE */

/*****************************************************************
 * dealing with other children
 */

#ifndef NO_OTHER_CHILD
API_EXPORT(void) ap_register_other_child(int pid,
		       void (*maintenance) (int reason, void *, int status),
			  void *data, int write_fd)
{
    other_child_rec *ocr;

    ocr = ap_palloc(pconf, sizeof(*ocr));
    ocr->pid = pid;
    ocr->maintenance = maintenance;
    ocr->data = data;
    ocr->write_fd = write_fd;
    ocr->next = other_children;
    other_children = ocr;
}

/* note that since this can be called by a maintenance function while we're
 * scanning the other_children list, all scanners should protect themself
 * by loading ocr->next before calling any maintenance function.
 */
API_EXPORT(void) ap_unregister_other_child(void *data)
{
    other_child_rec **pocr, *nocr;

    for (pocr = &other_children; *pocr; pocr = &(*pocr)->next) {
	if ((*pocr)->data == data) {
	    nocr = (*pocr)->next;
	    (*(*pocr)->maintenance) (OC_REASON_UNREGISTER, (*pocr)->data, -1);
	    *pocr = nocr;
	    /* XXX: um, well we've just wasted some space in pconf ? */
	    return;
	}
    }
}

/* test to ensure that the write_fds are all still writable, otherwise
 * invoke the maintenance functions as appropriate */
static void probe_writable_fds(void)
{
    /* XXX: this needs to be ported to NSPR */
    return;
#if 0
    fd_set writable_fds;
    int fd_max;
    other_child_rec *ocr, *nocr;
    struct timeval tv;
    int rc;

    if (other_children == NULL)
	return;

    fd_max = 0;
    FD_ZERO(&writable_fds);
    for (ocr = other_children; ocr; ocr = ocr->next) {
	if (ocr->write_fd == -1)
	    continue;
	FD_SET(ocr->write_fd, &writable_fds);
	if (ocr->write_fd > fd_max) {
	    fd_max = ocr->write_fd;
	}
    }
    if (fd_max == 0)
	return;

    do {
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	rc = ap_select(fd_max + 1, NULL, &writable_fds, NULL, &tv);
    } while (rc == -1 && errno == EINTR);

    if (rc == -1) {
	/* XXX: uhh this could be really bad, we could have a bad file
	 * descriptor due to a bug in one of the maintenance routines */
	ap_log_unixerr("probe_writable_fds", "select",
		    "could not probe writable fds", server_conf);
	return;
    }
    if (rc == 0)
	return;

    for (ocr = other_children; ocr; ocr = nocr) {
	nocr = ocr->next;
	if (ocr->write_fd == -1)
	    continue;
	if (FD_ISSET(ocr->write_fd, &writable_fds))
	    continue;
	(*ocr->maintenance) (OC_REASON_UNWRITABLE, ocr->data, -1);
    }
#endif
}

/* possibly reap an other_child, return 0 if yes, -1 if not */
static int reap_other_child(int pid, int status)
{
    other_child_rec *ocr, *nocr;

    for (ocr = other_children; ocr; ocr = nocr) {
	nocr = ocr->next;
	if (ocr->pid != pid)
	    continue;
	ocr->pid = -1;
	(*ocr->maintenance) (OC_REASON_DEATH, ocr->data, status);
	return 0;
    }
    return -1;
}
#endif

/*****************************************************************
 *
 * Dealing with the scoreboard... a lot of these variables are global
 * only to avoid getting clobbered by the longjmp() that happens when
 * a hard timeout expires...
 *
 * We begin with routines which deal with the file itself... 
 */

static void reinit_scoreboard(pool *p)
{
    ap_assert(!ap_scoreboard_image);
    ap_scoreboard_image = (scoreboard *) malloc(SCOREBOARD_SIZE);
    memset(ap_scoreboard_image, 0, SCOREBOARD_SIZE);
}

static void reopen_scoreboard(pool *p)
{
}


API_EXPORT(int) ap_exists_scoreboard_image(void)
{
    return (ap_scoreboard_image ? 1 : 0);
}

static ap_inline void put_scoreboard_info(int child_num,
				       short_score *new_score_rec)
{
}

/* a clean exit from the parent with proper cleanup */
static void __attribute__((noreturn)) clean_parent_exit(int code)
{
    /* Clear the pool - including any registered cleanups */
    ap_destroy_pool(pconf);
    exit(code);
}

int ap_update_child_status(int child_num, int status, request_rec *r)
{
    int old_status;
    short_score *ss;

    if (child_num < 0)
	return -1;

    ss = &ap_scoreboard_image->servers[child_num];
    old_status = ss->status;
    ss->status = status;
    ++ss->cur_vtime;

#if defined(STATUS)
    if (status == SERVER_READY || status == SERVER_DEAD) {
	/*
	 * Reset individual counters
	 */
	if (status == SERVER_DEAD) {
	    ss->my_access_count = 0L;
	    ss->my_bytes_served = 0L;
	}
	ss->conn_count = (unsigned short) 0;
	ss->conn_bytes = (unsigned long) 0;
    }
    if (r) {
	conn_rec *c = r->connection;
	ap_cpystrn(ss->client, ap_get_remote_host(c, r->per_dir_config,
			      REMOTE_NOLOOKUP), sizeof(ss->client));
	ap_cpystrn(ss->request, (r->the_request ? r->the_request :
			      "NULL"), sizeof(ss->request));
	ap_cpystrn(ss->vhost, r->server->server_hostname, sizeof(ss->vhost));
    }
#endif

    put_scoreboard_info(child_num, ss);

    return old_status;
}

static void update_scoreboard_global(void)
{
}

#if defined(STATUS)
void ap_time_process_request(int child_num, int status)
{
    short_score *ss;
#if defined(NO_GETTIMEOFDAY) && !defined(NO_TIMES)
    struct tms tms_blk;
#endif

    if (child_num < 0)
	return;

    ss = &ap_scoreboard_image->servers[child_num];

    if (status == START_PREQUEST) {
#if defined(NO_GETTIMEOFDAY)
#ifndef NO_TIMES
	if ((ss->start_time = times(&tms_blk)) == -1)
#endif /* NO_TIMES */
	    ss->start_time = (clock_t) 0;
#else
	if (gettimeofday(&ss->start_time, (struct timezone *) 0) < 0)
	    ss->start_time.tv_sec =
		ss->start_time.tv_usec = 0L;
#endif
    }
    else if (status == STOP_PREQUEST) {
#if defined(NO_GETTIMEOFDAY)
#ifndef NO_TIMES
	if ((ss->stop_time = times(&tms_blk)) == -1)
#endif
	    ss->stop_time = ss->start_time = (clock_t) 0;
#else
	if (gettimeofday(&ss->stop_time, (struct timezone *) 0) < 0)
	    ss->stop_time.tv_sec =
		ss->stop_time.tv_usec =
		ss->start_time.tv_sec =
		ss->start_time.tv_usec = 0L;
#endif

    }

    put_scoreboard_info(child_num, ss);
}

static void increment_counts(int child_num, request_rec *r)
{
    PRUint32 bs = 0;
    short_score *ss;

    ss = &ap_scoreboard_image->servers[child_num];

    if (r->sent_bodyct)
	ap_bgetopt(r->connection->client, BO_BYTECT, &bs);

#ifndef NO_TIMES
    times(&ss->times);
#endif
    ss->access_count++;
    ss->my_access_count++;
    ss->conn_count++;
    ss->bytes_served += (unsigned long) bs;
    ss->my_bytes_served += (unsigned long) bs;
    ss->conn_bytes += (unsigned long) bs;

    put_scoreboard_info(child_num, ss);
}

#endif


static int find_child_by_tid(PRThread *tid)
{
    int i;

    for (i = 0; i < max_daemons_limit; ++i)
	if (ap_scoreboard_image->parent[i].tid == tid)
	    return i;

    return -1;
}

static void reclaim_child_processes(int terminate)
{
    /* XXX: not necessary for NSPR */
}


/* Finally, this routine is used by the caretaker process to wait for
 * a while...
 */

/*****************************************************************
 * Connection structures and accounting...
 */

static void just_die(int sig)
{				/* SIGHUP to child process??? */
    /* if alarms are blocked we have to wait to die otherwise we might
     * end up with corruption in alloc.c's internal structures */
    if (tls()->alarms_blocked) {
	tls()->exit_after_unblock = 1;
    }
    else {
	clean_child_exit(0);
    }
}

/* volatile just in case */
static int volatile shutdown_pending;
static int volatile restart_pending;
static int volatile is_graceful;
static int volatile generation;


/*
 * start_shutdown() and start_restart(), below, are a first stab at
 * functions to initiate shutdown or restart without relying on signals. 
 * Previously this was initiated in sig_term() and restart() signal handlers, 
 * but we want to be able to start a shutdown/restart from other sources --
 * e.g. on Win32, from the service manager. Now the service manager can
 * call start_shutdown() or start_restart() as appropiate. 
 *
 * These should only be called from the parent process itself, since the
 * parent process will use the shutdown_pending and restart_pending variables
 * to determine whether to shutdown or restart. The child process should
 * call signal_parent() directly to tell the parent to die -- this will
 * cause neither of those variable to be set, which the parent will
 * assume means something serious is wrong (which it will be, for the
 * child to force an exit) and so do an exit anyway.
 */

void ap_start_shutdown(void)
{
    if (shutdown_pending == 1) {
	/* Um, is this _probably_ not an error, if the user has
	 * tried to do a shutdown twice quickly, so we won't
	 * worry about reporting it.
	 */
	return;
    }
    shutdown_pending = 1;
}

/* do a graceful restart if graceful == 1 */
void ap_start_restart(int graceful)
{
    if (restart_pending == 1) {
	/* Probably not an error - don't bother reporting it */
	return;
    }
    restart_pending = 1;
    is_graceful = graceful;
}

static void set_signals(void)
{
    /* XXX: what signals do we need to configure? */
}


/*****************************************************************
 * Here follows a long bunch of generic server bookkeeping stuff...
 */

/* check to see if we have the 'suexec' setuid wrapper installed */
static int init_suexec(void)
{
/* XXX: how do we do suexec test with NSPR? */
#if 0
    struct stat wrapper;

    if ((stat(SUEXEC_BIN, &wrapper)) != 0)
	return (ap_suexec_enabled);

    if ((wrapper.st_mode & S_ISUID) && wrapper.st_uid == 0) {
	ap_suexec_enabled = 1;
	PR_fprintf(PR_STDERR, "Configuring Apache for use with suexec wrapper.\n");
    }
    return (ap_suexec_enabled);
#else
    return 0;
#endif
}

/*****************************************************************
 * Connection structures and accounting...
 */


static conn_rec *new_connection(pool *p, server_rec *server, BUFF *inout,
			     const PRNetAddr *remaddr, const PRNetAddr *saddr,
			     int child_num)
{
    char buf[MAX_STRING_LEN];
    conn_rec *conn = (conn_rec *) ap_pcalloc(p, sizeof(conn_rec));

    /* Got a connection structure, so initialize what fields we can
     * (the rest are zeroed out by pcalloc).
     */

    conn->child_num = child_num;

    conn->pool = p;
    conn->local_addr = *saddr;
    conn->server = server; /* just a guess for now */
    ap_update_vhost_given_ip(conn);
    conn->base_server = conn->server;
    conn->client = inout;

    conn->remote_addr = *remaddr;
    if (PR_NetAddrToString(&conn->remote_addr, buf, sizeof(buf)) == PR_SUCCESS) {
	conn->remote_ip = ap_pstrdup(conn->pool, buf);
    }
    else {
	conn->remote_ip = "PR_NetAddrToString.error";
    }

    return conn;
}


static void sock_disable_nagle(PRFileDesc *s)
{
    /* The Nagle algorithm says that we should delay sending partial
     * packets in hopes of getting more data.  We don't want to do
     * this; we are not telnet.  There are bad interactions between
     * persistent connections and Nagle's algorithm that have very severe
     * performance penalties.  (Failing to disable Nagle is not much of a
     * problem with simple HTTP.)
     *
     * In spite of these problems, failure here is not a shooting offense.
     */
    PRSocketOptionData sod;

    sod.option = PR_SockOpt_NoDelay;
    sod.value.no_delay = 1;
    if (PR_SetSocketOption(s, &sod) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf,
		    "setsockopt: (TCP_NODELAY)");
    }
}


static PRFileDesc *make_sock(pool *p, const PRNetAddr *server)
{
    PRFileDesc *s;
    PRSocketOptionData sod;
    char addr[512];

    strcpy(addr, "address ");
    PR_NetAddrToString(server, addr + 8, sizeof(addr) - 8);

    /* note that because we're about to slack we don't use psocket */
    if ((s = ap_ptcpsocket(p)) == NULL) {
	ap_log_error(APLOG_MARK, APLOG_CRIT, server_conf,
		    "make_sock: failed to get a socket for %s", addr);
	exit(1);
    }

    sod.option = PR_SockOpt_Reuseaddr;
    sod.value.reuse_addr = 1;
    if (PR_SetSocketOption(s, &sod) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_CRIT, server_conf,
		    "make_sock: for %s, setsockopt: (SO_REUSEADDR)", addr);
	ap_pclosefd(p, s);
	return NULL;
    }
    sod.option = PR_SockOpt_Keepalive;
    sod.value.keep_alive = 1;
    if (PR_SetSocketOption(s, &sod) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_CRIT, server_conf,
		    "make_sock: for %s, setsockopt: (SO_KEEPALIVE)", addr);
	ap_pclosefd(p, s);
	return NULL;
    }

    sock_disable_nagle(s);
    sock_enable_linger(s);

    /*
     * To send data over high bandwidth-delay connections at full
     * speed we must force the TCP window to open wide enough to keep the
     * pipe full.  The default window size on many systems
     * is only 4kB.  Cross-country WAN connections of 100ms
     * at 1Mb/s are not impossible for well connected sites.
     * If we assume 100ms cross-country latency,
     * a 4kB buffer limits throughput to 40kB/s.
     *
     * To avoid this problem I've added the SendBufferSize directive
     * to allow the web master to configure send buffer size.
     *
     * The trade-off of larger buffers is that more kernel memory
     * is consumed.  YMMV, know your customers and your network!
     *
     * -John Heidemann <johnh@isi.edu> 25-Oct-96
     *
     * If no size is specified, use the kernel default.
     */
    if (server_conf->send_buffer_size) {
	sod.option = PR_SockOpt_SendBufferSize;
	sod.value.send_buffer_size = server_conf->send_buffer_size;
	if (PR_SetSocketOption(s, &sod) != PR_SUCCESS) {
	    ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf,
			"make_sock: failed to set SendBufferSize for %s, "
			"using default", addr);
	    /* not a fatal error */
	}
    }

    if (PR_Bind(s, server) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_CRIT, server_conf,
	    "make_sock: could not bind to %s", addr);
	ap_pclosefd(p, s);
	exit(1);
    }

    if (PR_Listen(s, ap_listenbacklog) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_ERR, server_conf,
	    "make_sock: unable to listen for connections on %s", addr);
	ap_pclosefd(p, s);
	exit(1);
    }

    return s;
}


/*
 * During a restart we keep track of the old listeners here, so that we
 * can re-use the sockets.  We have to do this because we won't be able
 * to re-open the sockets ("Address already in use").
 *
 * Unlike the listeners ring, old_listeners is a NULL terminated list.
 *
 * copy_listeners() makes the copy, find_listener() finds an old listener
 * and close_unused_listener() cleans up whatever wasn't used.
 */
static listen_rec *old_listeners;

/* unfortunately copy_listeners may be called before listeners is a ring */
static void copy_listeners(pool *p)
{
    listen_rec *lr;

    ap_assert(old_listeners == NULL);
    if (ap_listeners == NULL) {
	return;
    }
    lr = ap_listeners;
    do {
	listen_rec *nr = malloc(sizeof *nr);
	if (nr == NULL) {
	    PR_fprintf(PR_STDERR, "Ouch!  malloc failed in copy_listeners()\n");
	    exit(1);
	}
	*nr = *lr;
	ap_kill_cleanups_for_fd(p, nr->fd);
	nr->next = old_listeners;
	nr->index = 0;
	old_listeners = nr;
	lr = lr->next;
    } while (lr && lr != ap_listeners);
}


static PRFileDesc *find_listener(listen_rec *lr)
{
    listen_rec *or;

    for (or = old_listeners; or; or = or->next) {
	if (!memcmp(&or->local_addr, &lr->local_addr, sizeof(or->local_addr))) {
	    or->index = 1;
	    return or->fd;
	}
    }
    return NULL;
}


static void close_unused_listeners(void)
{
    listen_rec *or, *next;

    for (or = old_listeners; or; or = next) {
	next = or->next;
	if (!or->index)
	    PR_Close(or->fd);
	free(or);
    }
    old_listeners = NULL;
}


/* open sockets, and turn the listeners list into a singly linked ring */
static void setup_listeners(pool *p)
{
    listen_rec *lr;
    PRFileDesc *fd;
    PRIntn i;

    lr = ap_listeners;
    num_listenfds = 0;
    for (;;) {
	fd = find_listener(lr);
	if (fd == NULL) {
	    fd = make_sock(p, &lr->local_addr);
	}
	else {
	    ap_note_cleanups_for_fd(p, fd);
	}
	if (fd) {
	    ++num_listenfds;
	}
	lr->fd = fd;
	if (lr->next == NULL)
	    break;
	lr = lr->next;
    }
    /* turn the list into a ring */
    lr->next = ap_listeners;
    close_unused_listeners();

    listenfds = ap_palloc(p, sizeof(PRPollDesc) * num_listenfds);
    for (lr = ap_listeners, i = 0; i < num_listenfds; lr = lr->next, ++i) {
	lr->index = i;
	listenfds[i].fd = lr->fd;
	listenfds[i].in_flags = PR_POLL_READ;
	listenfds[i].out_flags = 0;
    }

#ifdef NO_SERIALIZED_ACCEPT
    /* warn them about the starvation problem if they're using multiple
     * sockets
     */
    if (num_listenfds > 1) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, NULL,
		    "You cannot use multiple Listens safely on your system, "
		    "proceeding anyway.  See src/PORTING, search for "
		    "SERIALIZED_ACCEPT.");
    }
#endif
}


static void show_compile_settings(void)
{
#ifdef SERVER_SUBVERSION
    printf("Server base version: %s\n", SERVER_BASEVERSION);
    printf("Server sub-version:  %s\n", SERVER_SUBVERSION);
    printf("Server built:        %s\n", ap_get_server_built());
#else
    printf("Server version: %s\n", ap_get_server_version());
    printf("Server built:   %s\n", ap_get_server_built());
#endif
    printf("Server's Module Magic Number: %u\n", MODULE_MAGIC_NUMBER);
    printf("Server compiled with....\n");
#ifdef BIG_SECURITY_HOLE
    printf(" -D BIG_SECURITY_HOLE\n");
#endif
#ifdef SECURITY_HOLE_PASS_AUTHORIZATION
    printf(" -D SECURITY_HOLE_PASS_AUTHORIZATION\n");
#endif
#ifdef HTTPD_ROOT
    printf(" -D HTTPD_ROOT=\"" HTTPD_ROOT "\"\n");
#endif
#ifdef HAVE_MMAP
    printf(" -D HAVE_MMAP\n");
#endif
#ifdef HAVE_SHMGET
    printf(" -D HAVE_SHMGET\n");
#endif
#ifdef USE_MMAP_SCOREBOARD
    printf(" -D USE_MMAP_SCOREBOARD\n");
#endif
#ifdef USE_SHMGET_SCOREBOARD
    printf(" -D USE_SHMGET_SCOREBOARD\n");
#endif
#ifdef USE_OS2_SCOREBOARD
    printf(" -D USE_OS2_SCOREBOARD\n");
#endif
#ifdef USE_POSIX_SCOREBOARD
    printf(" -D USE_POSIX_SCOREBOARD\n");
#endif
#ifdef USE_MMAP_FILES
    printf(" -D USE_MMAP_FILES\n");
#ifdef MMAP_SEGMENT_SIZE
	printf(" -D MMAP_SEGMENT_SIZE=%ld\n",(long)MMAP_SEGMENT_SIZE);
#endif
#endif /*USE_MMAP_FILES*/
#ifdef NO_WRITEV
    printf(" -D NO_WRITEV\n");
#endif
#ifdef NO_LINGCLOSE
    printf(" -D NO_LINGCLOSE\n");
#endif
#ifdef SINGLE_LISTEN_UNSERIALIZED_ACCEPT
    printf(" -D SINGLE_LISTEN_UNSERIALIZED_ACCEPT\n");
#endif
#ifdef NO_OTHER_CHILD
    printf(" -D NO_OTHER_CHILD\n");
#endif
#ifdef NO_RELIABLE_PIPED_LOGS
    printf(" -D NO_RELIABLE_PIPED_LOGS\n");
#endif
#ifdef BUFFERED_LOGS
    printf(" -D BUFFERED_LOGS\n");
#ifdef PIPE_BUF
	printf(" -D PIPE_BUF=%ld\n",(long)PIPE_BUF);
#endif
#endif
#ifdef CHARSET_EBCDIC
    printf(" -D CHARSET_EBCDIC\n");
#endif
#ifdef NEED_HASHBANG_EMUL
    printf(" -D NEED_HASHBANG_EMUL\n");
#endif
#ifdef SHARED_CORE
    printf(" -D SHARED_CORE\n");
#endif
}


/* Some init code that's common between win32 and unix... well actually
 * some of it is #ifdef'd but was duplicated before anyhow.  This stuff
 * is still a mess.
 */
static void common_init(void)
{
#ifdef AUX3
    (void) set42sig();
#endif

    pperm = ap_init_alloc();
    pconf = ap_make_sub_pool(pperm);
    ptemp = ap_make_sub_pool(pconf);

    ap_util_init();
    ap_util_uri_init();

    pcommands = ap_make_sub_pool(pperm);
    ap_server_pre_read_config  = ap_make_array(pcommands, 1, sizeof(char *));
    ap_server_post_read_config = ap_make_array(pcommands, 1, sizeof(char *));
}

/*****************************************************************
 * Child process main loop.
 */


API_EXPORT(void) ap_child_terminate(request_rec *r)
{
    r->connection->keepalive = 0;
    tls()->requests_this_child = ap_max_requests_per_child = 1;
}

static PRFileDesc *accept_loop(PRNetAddr *sa_client)
{
    PRInt32 srv;
    listen_rec *lr;
    PRFileDesc *sd;
    PRFileDesc *csd;

    PR_Lock(accept_loop_lock);
    for (;;) {
	if (num_listenfds > 1) {
	    /* more than one socket */
	    srv = PR_Poll(listenfds, num_listenfds, PR_INTERVAL_NO_TIMEOUT);
	    if (srv < 0) {
		PR_Unlock(accept_loop_lock);
		/* Single Unix documents select as returning errnos
		 * EBADF, EINTR, and EINVAL... and in none of those
		 * cases does it make sense to continue.  In fact
		 * on Linux 2.0.x we seem to end up with EFAULT
		 * occasionally, and we'd loop forever due to it.
		 */
		ap_log_error(APLOG_MARK, APLOG_ERR, server_conf, "select: (listen)");
		clean_child_exit(1);
	    }

	    if (srv == 0)
		continue;

	    /* find a listener */
	    lr = head_listener;
	    do {
		/* XXX: should we check for PR_POLL_ERR ?? */
		if (listenfds[lr->index].out_flags & PR_POLL_READ) {
		    /* advance to the next listener for next loop */
		    head_listener = lr->next;
		    goto got_lr;
		}
		lr = lr->next;
	    } while (lr != head_listener);

	    /* if we don't find anything then just start again */
	    continue;

got_lr:
	    sd = lr->fd;
	}
	else {
	    /* only one socket, just pretend we did the other stuff */
	    sd = ap_listeners->fd;
	}

	csd = PR_Accept(sd, sa_client, PR_INTERVAL_NO_TIMEOUT);

	if (csd)
	    break;		/* We have a socket ready for reading */
	/* XXX: we need to deal with error conditions here */
    }
    PR_Unlock(accept_loop_lock);
    return csd;
}

static void child_main(int child_num_arg, tls_main_t *t)
{
    PRNetAddr sa_server;
    PRNetAddr sa_client;
    pool *ptrans;
    PRFileDesc *csd;
    conn_rec *current_conn;

    /* All of initialization is a critical section, we don't care if we're
     * told to HUP or USR1 before we're done initializing.  For example,
     * we could be half way through child_init_modules() when a restart
     * signal arrives, and we'd have no real way to recover gracefully
     * and exit properly.
     *
     * I suppose a module could take forever to initialize, but that would
     * be either a broken module, or a broken configuration (i.e. network
     * problems, file locking problems, whatever). -djg
     */

    csd = NULL;
    t->my_child_num = child_num_arg;
    t->requests_this_child = 0;

    /* Get a sub pool for global allocations in this child, so that
     * we can have cleanups occur when the child exits.
     */
    t->pchild = ap_make_root_pool();

    /* needs to be done before we switch UIDs so we have permissions */
    reopen_scoreboard(t->pchild);
    SAFE_ACCEPT(accept_mutex_child_init(t->pchild));

    /* XXX: we need to set_group_privs and setuid and such for NSPR */
#if 0
    set_group_privs();
    /* Only try to switch if we're running as root */
    if (!geteuid() && setuid(ap_user_id) == -1) {
	ap_log_error(APLOG_MARK, APLOG_ALERT, server_conf,
		    "setuid: unable to change uid");
	clean_child_exit(1);
    }
#endif

    ap_child_init_modules(t->pchild, server_conf);

    (void) ap_update_child_status(t->my_child_num, SERVER_READY, (request_rec *) NULL);

    /*
     * Setup the jump buffers so that we can return here after
     * a signal or a timeout (yeah, I know, same thing).
     */
    setjmp(t->jmpbuffer);

    ptrans = ap_make_sub_pool(t->pchild);

    while (1) {
	BUFF *conn_io;
	request_rec *r;

	/*
	 * (Re)initialize this child to a pre-connection state.
	 */

	current_conn = NULL;

	ap_clear_pool(ptrans);

	if (ap_scoreboard_image->global.exit_generation >= generation) {
	    clean_child_exit(0);
	}

	if ((ap_max_requests_per_child > 0
	     && ++t->requests_this_child >= ap_max_requests_per_child)) {
	    clean_child_exit(0);
	}

	(void) ap_update_child_status(t->my_child_num, SERVER_READY, (request_rec *) NULL);

	/*
	 * Wait for an acceptable connection to arrive.
	 */
	csd = accept_loop(&sa_client);

	ap_note_cleanups_for_fd(ptrans, csd);

	/*
	 * We now have a connection, so set it up with the appropriate
	 * socket options, file descriptors, and read/write buffers.
	 */

	if (PR_GetSockName(csd, &sa_server) != PR_SUCCESS) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, server_conf, "PR_GetSockName");
	    continue;
	}

	sock_disable_nagle(csd);

	(void) ap_update_child_status(t->my_child_num, SERVER_BUSY_READ,
				   (request_rec *) NULL);

	conn_io = ap_bcreate(ptrans, B_RDWR | B_SOCKET);
	ap_bpushfd(conn_io, csd);

	current_conn = new_connection(ptrans, server_conf, conn_io,
				    &sa_client, &sa_server, t->my_child_num);

	/*
	 * Read and process each request found on our connection
	 * until no requests are left or we decide to close.
	 */

	while ((r = ap_read_request(current_conn)) != NULL) {

	    /* read_request_line has already done a
	     * signal (SIGUSR1, SIG_IGN);
	     */

	    (void) ap_update_child_status(t->my_child_num, SERVER_BUSY_WRITE, r);

	    ap_process_request(r);

#if defined(STATUS)
	    increment_counts(t->my_child_num, r);
#endif

	    if (!current_conn->keepalive || ap_is_aborted(current_conn))
		break;

	    ap_destroy_pool(r->pool);
	    (void) ap_update_child_status(t->my_child_num, SERVER_BUSY_KEEPALIVE,
				       (request_rec *) NULL);

	    if (ap_scoreboard_image->global.exit_generation >= generation) {
		ap_bclose(conn_io);
		clean_child_exit(0);
	    }
	}

	/*
	 * Close the connection, being careful to send out whatever is still
	 * in our buffers.  If possible, try to avoid a hard close until the
	 * client has ACKed our FIN and/or has stopped sending us data.
	 */

#ifdef NO_LINGCLOSE
	ap_bclose(conn_io);	/* just close it */
#else
	if (r && r->connection
	    && !ap_is_aborted(r->connection)
	    && r->connection->client
	    && (r->connection->client->fd >= 0)) {

	    lingering_close(r);
	}
	else {
	    ap_bsetflag(conn_io, B_EOUT, 1);
	    ap_bclose(conn_io);
	}
#endif
    }
}


static tls_main_t *create_tls(void)
{
    tls_main_t *t;
    PRStatus rc;

    t = malloc(sizeof(*t));
    memset(t, 0, sizeof(*t));
    if ((rc = PR_SetThreadPrivate(tls_main_key, t)) != PR_SUCCESS) {
	PR_fprintf(PR_STDERR, "PR_SetThreadPrivate error\n");
	exit(1);
    }

    return t;
}


static void PR_CALLBACK child_starter(void *dummy)
{
    tls_main_t *t;

    t = create_tls();
    t->usr1_just_die = 1;

    if (!setjmp(t->thread_exit)) {
	child_main((int)dummy, t);
    }
}

static int make_child(server_rec *s, int slot, PRTime now)
{
    PRThread *tid;

    if (slot + 1 > max_daemons_limit) {
	max_daemons_limit = slot + 1;
    }

    if (one_process) {
	child_main(slot, tls());
    }

    (void) ap_update_child_status(slot, SERVER_STARTING, (request_rec *) NULL);

    tid = PR_CreateThread(PR_USER_THREAD, child_starter, (void *)slot,
	    PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD, PR_UNJOINABLE_THREAD, 0);
    if (tid == NULL) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, s,
	    "PR_CreateThread failed");

	/* fork didn't succeed. Fix the scoreboard or else
	 * it will say SERVER_STARTING forever and ever
	 */
	(void) ap_update_child_status(slot, SERVER_DEAD, (request_rec *) NULL);

	/* In case system resources are maxxed out, we don't want
	   Apache running away with the CPU trying to fork over and
	   over and over again. */
	PR_Sleep(PR_SecondsToInterval(10));

	return -1;
    }

    ap_scoreboard_image->parent[slot].last_rtime = now;
    ap_scoreboard_image->parent[slot].tid = tid;

    return 0;
}



/* start up a bunch of children */
static void startup_children(int number_to_start)
{
    int i;
    PRTime now = PR_Now();

    for (i = 0; number_to_start && i < ap_daemons_limit; ++i) {
	if (ap_scoreboard_image->servers[i].status != SERVER_DEAD) {
	    continue;
	}
	if (make_child(server_conf, i, now) < 0) {
	    break;
	}
	--number_to_start;
    }
}


/*
 * idle_spawn_rate is the number of children that will be spawned on the
 * next maintenance cycle if there aren't enough idle servers.  It is
 * doubled up to MAX_SPAWN_RATE, and reset only when a cycle goes by
 * without the need to spawn.
 */
static int idle_spawn_rate = 1;
#ifndef MAX_SPAWN_RATE
#define MAX_SPAWN_RATE	(32)
#endif
static int hold_off_on_exponential_spawning;

static void perform_idle_server_maintenance(void)
{
    int i;
    int to_kill;
    int idle_count;
    short_score *ss;
    PRIntervalTime inow = PR_IntervalNow();
    int free_length;
    int free_slots[MAX_SPAWN_RATE];
    int last_non_dead;
    int total_non_dead;

    /* initialize the free_list */
    free_length = 0;

    to_kill = -1;
    idle_count = 0;
    last_non_dead = -1;
    total_non_dead = 0;

    for (i = 0; i < ap_daemons_limit; ++i) {
	int status;

	if (i >= max_daemons_limit && free_length == idle_spawn_rate)
	    break;
	ss = &ap_scoreboard_image->servers[i];
	status = ss->status;
	if (status == SERVER_DEAD) {
	    /* try to keep children numbers as low as possible */
	    if (free_length < idle_spawn_rate) {
		free_slots[free_length] = i;
		++free_length;
	    }
	}
	else {
	    /* We consider a starting server as idle because we started it
	     * at least a cycle ago, and if it still hasn't finished starting
	     * then we're just going to swamp things worse by forking more.
	     * So we hopefully won't need to fork more if we count it.
	     * This depends on the ordering of SERVER_READY and SERVER_STARTING.
	     */
	    if (status <= SERVER_READY) {
		++ idle_count;
		/* always kill the highest numbered child if we have to...
		 * no really well thought out reason ... other than observing
		 * the server behaviour under linux where lower numbered children
		 * tend to service more hits (and hence are more likely to have
		 * their data in cpu caches).
		 */
		to_kill = i;
	    }

	    ++total_non_dead;
	    last_non_dead = i;
	    if (ss->timeout_len) {
		/* if it's a live server, with a live timeout then
		 * start checking its timeout */
		parent_score *ps = &ap_scoreboard_image->parent[i];
		if (ss->cur_vtime != ps->last_vtime) {
		    /* it has made progress, so update its last_rtime,
		     * last_vtime */
		    ps->last_rtime = inow;
		    ps->last_vtime = ss->cur_vtime;
		}
		else if ((PRIntervalTime)(inow - ps->last_rtime) > ss->timeout_len) {
		    /* no progress, and the timeout length has been exceeded */
		    ss->timeout_len = 0;
		    /* XXX: NSPR timeouts */
		    /* kill(ps->pid, SIGALRM); */
		}
	    }
	}
    }
    max_daemons_limit = last_non_dead + 1;
    if (idle_count > ap_daemons_max_free) {
	/* kill off one child... we use SIGUSR1 because that'll cause it to
	 * shut down gracefully, in case it happened to pick up a request
	 * while we were counting
	 */
	/* XXX: NSPR idle maintenance */
	/* kill(ap_scoreboard_image->parent[to_kill].pid, SIGUSR1); */
	idle_spawn_rate = 1;
    }
    else if (idle_count < ap_daemons_min_free) {
	/* terminate the free list */
	if (free_length == 0) {
	    /* only report this condition once */
	    static int reported = 0;

	    if (!reported) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, server_conf,
			    "server reached MaxClients setting, consider"
			    " raising the MaxClients setting");
		reported = 1;
	    }
	    idle_spawn_rate = 1;
	}
	else {
	    PRTime now = PR_Now();

	    if (idle_spawn_rate >= 8) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, server_conf,
		    "server seems busy, (you may need "
		    "to increase StartServers, or Min/MaxSpareServers), "
		    "spawning %d children, there are %d idle, and "
		    "%d total children", idle_spawn_rate,
		    idle_count, total_non_dead);
	    }
	    for (i = 0; i < free_length; ++i) {
		make_child(server_conf, free_slots[i], now);
	    }
	    /* the next time around we want to spawn twice as many if this
	     * wasn't good enough, but not if we've just done a graceful
	     */
	    if (hold_off_on_exponential_spawning) {
		--hold_off_on_exponential_spawning;
	    }
	    else if (idle_spawn_rate < MAX_SPAWN_RATE) {
		idle_spawn_rate *= 2;
	    }
	}
    }
    else {
	idle_spawn_rate = 1;
    }
}


/*****************************************************************
 * Executive routines.
 */

static void server_main_loop(int remaining_children_to_start)
{
    int child_slot;
    int status;
    int pid;

    head_listener = ap_listeners;

    while (!restart_pending && !shutdown_pending) {
	PR_Sleep(PR_IntervalToSeconds(10000));
	pid = -1;
	/* XXX: if it takes longer than 1 second for all our children
	    * to start up and get into IDLE state then we may spawn an
	    * extra child
	    */
	if (pid >= 0) {
	    /* Child died... note that it's gone in the scoreboard. */
	    child_slot = find_child_by_tid(NULL);
	    Explain2("Reaping child %d slot %d", pid, child_slot);
	    if (child_slot >= 0) {
		(void) ap_update_child_status(child_slot, SERVER_DEAD,
					    (request_rec *) NULL);
		if (remaining_children_to_start
		    && child_slot < ap_daemons_limit) {
		    /* we're still doing a 1-for-1 replacement of dead
			* children with new children
			*/
		    make_child(server_conf, child_slot, PR_Now());
		    --remaining_children_to_start;
		}
#ifndef NO_OTHER_CHILD
	    }
	    else if (reap_other_child(pid, status) == 0) {
		/* handled */
#endif
	    }
	    else if (is_graceful) {
		/* Great, we've probably just lost a slot in the
		    * scoreboard.  Somehow we don't know about this
		    * child.
		    */
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, server_conf,
			    "long lost child came home! (pid %d)", pid);
	    }
	    /* Don't perform idle maintenance when a child dies,
		* only do it when there's a timeout.  Remember only a
		* finite number of children can die, and it's pretty
		* pathological for a lot to die suddenly.
		*/
	    continue;
	}
	else if (remaining_children_to_start) {
	    /* we hit a 1 second timeout in which none of the previous
		* generation of children needed to be reaped... so assume
		* they're all done, and pick up the slack if any is left.
		*/
	    startup_children(remaining_children_to_start);
	    remaining_children_to_start = 0;
	    /* In any event we really shouldn't do the code below because
		* few of the servers we just started are in the IDLE state
		* yet, so we'd mistakenly create an extra server.
		*/
	    continue;
	}

	perform_idle_server_maintenance();
    }
}

static void one_config_cycle(void)
{
    int remaining_children_to_start;

    copy_listeners(pconf);
    if (!is_graceful) {
	ap_restart_time = PR_Now();
    }
    ap_clear_pool(pconf);
    ptemp = ap_make_sub_pool(pconf);

    server_conf = ap_read_config(pconf, ptemp, ap_server_confname);
    setup_listeners(pconf);
    ap_open_logs(server_conf, pconf);
    ap_log_pid(pconf, ap_pid_fname);
    ap_init_modules(pconf, server_conf);
    SAFE_ACCEPT(accept_mutex_init(pconf));
    if (!is_graceful) {
	reinit_scoreboard(pconf);
    }
    set_signals();

    /* set up accept_loop */
    head_listener = ap_listeners;
    accept_loop_lock = PR_NewLock();

    if (ap_daemons_max_free < ap_daemons_min_free + 1)	/* Don't thrash... */
	ap_daemons_max_free = ap_daemons_min_free + 1;

    /* If we're doing a graceful_restart then we're going to see a lot
     * of children exiting immediately when we get into the main loop
     * below (because we just sent them SIGUSR1).  This happens pretty
     * rapidly... and for each one that exits we'll start a new one until
     * we reach at least daemons_min_free.  But we may be permitted to
     * start more than that, so we'll just keep track of how many we're
     * supposed to start up without the 1 second penalty between each fork.
     */
    remaining_children_to_start = ap_daemons_to_start;
    if (remaining_children_to_start > ap_daemons_limit) {
	remaining_children_to_start = ap_daemons_limit;
    }
    if (!is_graceful) {
	startup_children(remaining_children_to_start);
	remaining_children_to_start = 0;
    }
    else {
	/* give the system some time to recover before kicking into
	    * exponential mode */
	hold_off_on_exponential_spawning = 10;
    }

    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, server_conf,
		"%s configured -- resuming normal operations",
		ap_get_server_version());
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, server_conf,
		"Server built: %s", ap_get_server_built());
    restart_pending = shutdown_pending = 0;

    server_main_loop(remaining_children_to_start);

    if (shutdown_pending) {
	/* NSPR: shutdown code */
#if 0
	/* Time to gracefully shut down:
	    * Kill child processes, tell them to call child_exit, etc...
	    */
	if (ap_killpg(pgrp, SIGTERM) < 0) {
	    ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf, "killpg SIGTERM");
	}
	reclaim_child_processes(1);		/* Start with SIGTERM */
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, server_conf,
		    "httpd: caught SIGTERM, shutting down");
#endif
	clean_parent_exit(0);
    }

    /* we've been told to restart */

    if (one_process) {
	/* not worth thinking about */
	clean_parent_exit(0);
    }

    if (is_graceful) {
	int i;

	/* USE WITH CAUTION:  Graceful restarts are not known to work
	    * in various configurations on the architectures we support. */
	ap_scoreboard_image->global.exit_generation = generation;
	update_scoreboard_global();

	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, server_conf,
		    "SIGUSR1 received.  Doing graceful restart");

	/* kill off the idle ones */
#if 0
	if (ap_killpg(pgrp, SIGUSR1) < 0) {
	    ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf, "killpg SIGUSR1");
	}
#endif
	/* This is mostly for debugging... so that we know what is still
	    * gracefully dealing with existing request.
	    */
	for (i = 0; i < ap_daemons_limit; ++i) {
	    if (ap_scoreboard_image->servers[i].status != SERVER_DEAD) {
		ap_scoreboard_image->servers[i].status = SERVER_GRACEFUL;
	    }
	}
    }
    else {
	/* Kill 'em off */
#if 0
	if (ap_killpg(pgrp, SIGHUP) < 0) {
	    ap_log_error(APLOG_MARK, APLOG_WARNING, server_conf, "killpg SIGHUP");
	}
	reclaim_child_processes(0);		/* Not when just starting up */
#endif
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, server_conf,
		    "SIGHUP received.  Attempting to restart");
    }

    PR_DestroyLock(accept_loop_lock);
}

static void standalone_main(int argc, char **argv)
{
    ap_standalone = 1;

    is_graceful = 0;
    ++generation;

    if (!one_process) {
#if 0
	detach();
#endif
    }
    else {
	MONCONTROL(1);
    }

    do {
	one_config_cycle();
	++generation;
    } while (restart_pending);

    /*add_common_vars(NULL);*/
}				/* standalone_main */

extern char *optarg;
extern int optind;

/* XXX: for getopt... which we need to replace */
#include <unistd.h>

static PRIntn PR_CALLBACK nspr_main(PRIntn argc, char *argv[])
{
    int c;
    PRStatus rc;

    rc = PR_NewThreadPrivateIndex(&tls_main_key, NULL);
    if (rc != PR_SUCCESS) {
	PR_fprintf(PR_STDERR, "PR_NewThreadPrivateIndex failed\n");
	exit(1);
    }
    create_tls();

    MONCONTROL(0);

    common_init();
    
    ap_server_argv0 = argv[0];
    ap_cpystrn(ap_server_root, HTTPD_ROOT, sizeof(ap_server_root));
    ap_cpystrn(ap_server_confname, SERVER_CONFIG_FILE, sizeof(ap_server_confname));

    ap_setup_prelinked_modules();

    while ((c = getopt(argc, argv,
				    "C:c:Xd:f:vVhlL:S"
#ifdef DEBUG_SIGSTOP
				    "Z:"
#endif
			)) != -1) {
	char **new;
	switch (c) {
	case 'c':
	    new = (char **)ap_push_array(ap_server_post_read_config);
	    *new = ap_pstrdup(pcommands, optarg);
	    break;
	case 'C':
	    new = (char **)ap_push_array(ap_server_pre_read_config);
	    *new = ap_pstrdup(pcommands, optarg);
	    break;
	case 'd':
	    ap_cpystrn(ap_server_root, optarg, sizeof(ap_server_root));
	    break;
	case 'f':
	    ap_cpystrn(ap_server_confname, optarg, sizeof(ap_server_confname));
	    break;
	case 'v':
	    printf("Server version: %s\n", ap_get_server_version());
	    printf("Server built:   %s\n", ap_get_server_built());
	    exit(0);
	case 'V':
	    show_compile_settings();
	    exit(0);
	case 'h':
	    ap_show_directives();
	    exit(0);
	case 'l':
	    ap_show_modules();
	    exit(0);
	case 'X':
	    ++one_process;	/* Weird debugging mode. */
	    break;
#ifdef DEBUG_SIGSTOP
	case 'Z':
	    raise_sigstop_flags = atoi(optarg);
	    break;
#endif
#ifdef SHARED_CORE
	case 'L':
	    /* just ignore this option here, because it has only
	     * effect when SHARED_CORE is used and then it was
	     * already handled in the Shared Core Bootstrap
	     * program.
	     */
	    break;
#endif
	case 'S':
	    ap_dump_settings = 1;
	    break;
	case '?':
	    usage(argv[0]);
	}
    }

#ifdef __EMX__
    printf("%s \n", ap_get_server_version());
#endif

    ap_suexec_enabled = init_suexec();
    server_conf = ap_read_config(pconf, ptemp, ap_server_confname);

    child_timeouts = !ap_standalone || one_process;

    if (!ap_standalone) {
	PR_fprintf(PR_STDERR, "inetd mode not supported\n");
	exit(1);
    }

    ap_open_logs(server_conf, pconf);
    ap_init_modules(pconf, server_conf);
    standalone_main(argc, argv);
    exit(0);
}

int main(int argc, char **argv)
{
    return PR_Initialize(nspr_main, argc, argv, 0);
}
